/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *
 * This file is part of xlslib -- A multiplatform, C/C++ library
 * for dynamic generation of Excel(TM) files.
 *
 * Copyright 2004 Yeico S. A. de C. V. All Rights Reserved.
 * Copyright 2008-2011 David Hoerl All Rights Reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification, are
 * permitted provided that the following conditions are met:
 *
 *    1. Redistributions of source code must retain the above copyright notice, this list of
 *       conditions and the following disclaimer.
 *
 *    2. Redistributions in binary form must reproduce the above copyright notice, this list
 *       of conditions and the following disclaimer in the documentation and/or other materials
 *       provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY David Hoerl ''AS IS'' AND ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL David Hoerl OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 * ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#include "xlslib/merged.h"
#include "xlslib/common.h"
#include "xlslib/rectypes.h"


using namespace std;
using namespace xlslib_core;

/*
 ******************************
 * CMergedCells class implementation
 ******************************
 */
CMergedCells::CMergedCells(CDataStorage &datastore) :
	CRecord(datastore)
{
	SetRecordType(RECTYPE_MERGEDCELLS);

	// By default the record is empty
	AddValue16(0x00);

	SetRecordLength(GetDataSize()-RECORD_HEADER_SIZE);
}

CMergedCells::~CMergedCells()
{
}

void CMergedCells::AddRange(range_t* rng)
{
	AddValue16((unsigned16_t)rng->first_row);
	AddValue16((unsigned16_t)rng->last_row);
	AddValue16((unsigned16_t)rng->first_col);
	AddValue16((unsigned16_t)rng->last_col);

	SetRecordLength(GetDataSize()-RECORD_HEADER_SIZE);
}

void CMergedCells::SetNumRanges(size_t numranges)
{
	SetValueAt16((unsigned16_t)numranges, 0x04);
}
